@@ -105,11 +105,8 @@ def adjoint_tensor(adj_redop, adj_binop, out_adj, data, inputs, dtype):
 def adjoint_binary(adj_redop, adj_binop, out_adj, op, lhs, rhs):
     assert (adj_redop, op) in ops.DISTRIBUTIVE_OPS
 
-    lhs_reduced_vars = frozenset(rhs.inputs) - frozenset(lhs.inputs)
-    lhs_adj = op(out_adj, rhs).reduce(adj_redop, lhs_reduced_vars)
-
-    rhs_reduced_vars = frozenset(lhs.inputs) - frozenset(rhs.inputs)
-    rhs_adj = op(out_adj, lhs).reduce(adj_redop, rhs_reduced_vars)
+    lhs_adj = op(out_adj, rhs).reduce(adj_redop, rhs.input_vars - lhs.input_vars)
+    rhs_adj = op(out_adj, lhs).reduce(adj_redop, lhs.input_vars - rhs.input_vars)
 
     return {lhs: lhs_adj, rhs: rhs_adj}
 
@@ -144,11 +141,10 @@ def adjoint_contract_generic(adj_redop, adj_binop, out_adj, sum_op, prod_op, red
 def adjoint_contract(adj_redop, adj_binop, out_adj, sum_op, prod_op, reduced_vars, lhs, rhs):
     assert sum_op is nullop or (sum_op, prod_op) in ops.DISTRIBUTIVE_OPS
 
-    lhs_reduced_vars = frozenset(rhs.inputs) - frozenset(lhs.inputs)
-    lhs_adj = Contraction(sum_op if sum_op is not nullop else adj_redop, prod_op, lhs_reduced_vars, out_adj, rhs)
-
-    rhs_reduced_vars = frozenset(lhs.inputs) - frozenset(rhs.inputs)
-    rhs_adj = Contraction(sum_op if sum_op is not nullop else adj_redop, prod_op, rhs_reduced_vars, out_adj, lhs)
+    lhs_adj = Contraction(sum_op if sum_op is not nullop else adj_redop,
+                          prod_op, rhs.input_vars - lhs.input_vars, out_adj, rhs)
+    rhs_adj = Contraction(sum_op if sum_op is not nullop else adj_redop,
+                          prod_op, lhs.input_vars - rhs.input_vars, out_adj, lhs)
 
     return {lhs: lhs_adj, rhs: rhs_adj}
 
@@ -260,7 +256,7 @@ def _scatter(src, res, subs):
     return Tensor(data, inputs, res.dtype)
 
 
-@adjoint_ops.register(Subs, ops.LogAddExpOp, ops.AddOp, GaussianMixture, GaussianMixture, tuple)
+@adjoint_ops.register(Subs, ops.LogaddexpOp, ops.AddOp, GaussianMixture, GaussianMixture, tuple)
 def adjoint_subs_gaussianmixture_gaussianmixture(adj_redop, adj_binop, out_adj, arg, subs):
 
     if any(v.dtype == 'real' and not isinstance(v, Variable) for k, v in subs):
@@ -305,7 +301,7 @@ def adjoint_subs_gaussianmixture_gaussianmixture(adj_redop, adj_binop, out_adj,
     return {arg: in_adj}
 
 
-@adjoint_ops.register(Subs, ops.LogAddExpOp, ops.AddOp, Gaussian, GaussianMixture, tuple)
+@adjoint_ops.register(Subs, ops.LogaddexpOp, ops.AddOp, Gaussian, GaussianMixture, tuple)
 def adjoint_subs_gaussianmixture_discrete(adj_redop, adj_binop, out_adj, arg, subs):
 
     if any(v.dtype == 'real' and not isinstance(v, Variable) for k, v in subs):
@@ -316,7 +312,7 @@ def adjoint_subs_gaussianmixture_discrete(adj_redop, adj_binop, out_adj, arg, su
     return {arg: adjoint_ops(Subs, adj_redop, adj_binop, out_adj_, arg, subs)[arg]}
 
 
-@adjoint_ops.register(Subs, ops.LogAddExpOp, ops.AddOp, (GaussianMixture, Gaussian), Gaussian, tuple)
+@adjoint_ops.register(Subs, ops.LogaddexpOp, ops.AddOp, (GaussianMixture, Gaussian), Gaussian, tuple)
 def adjoint_subs_gaussian_gaussian(adj_redop, adj_binop, out_adj, arg, subs):
 
     if any(v.dtype == 'real' and not isinstance(v, Variable) for k, v in subs):
@@ -327,7 +323,7 @@ def adjoint_subs_gaussian_gaussian(adj_redop, adj_binop, out_adj, arg, subs):
     return {arg: adjoint_ops(Subs, adj_redop, adj_binop, out_adj, arg_, subs)[arg_]}
 
 
-@adjoint_ops.register(Subs, ops.LogAddExpOp, ops.AddOp, (Number, Tensor), GaussianMixture, tuple)
+@adjoint_ops.register(Subs, ops.LogaddexpOp, ops.AddOp, (Number, Tensor), GaussianMixture, tuple)
 def adjoint_subs_gaussianmixture_discrete(adj_redop, adj_binop, out_adj, arg, subs):
 
     if any(v.dtype == 'real' and not isinstance(v, Variable) for k, v in subs):
