@@ -3,13 +3,14 @@
 
 import functools
 import itertools
-from collections import OrderedDict, defaultdict
+from collections import Counter, OrderedDict, defaultdict
 from functools import reduce
 from typing import Tuple, Union
 
 import opt_einsum
 from multipledispatch.variadic import Variadic
 
+import funsor
 import funsor.ops as ops
 from funsor.affine import affine_inputs
 from funsor.delta import Delta
@@ -50,7 +51,7 @@ class Contraction(Funsor):
         assert isinstance(bin_op, AssociativeOp)
         assert all(isinstance(v, Funsor) for v in terms)
         assert isinstance(reduced_vars, frozenset)
-        assert all(isinstance(v, str) for v in reduced_vars)
+        assert all(isinstance(v, Variable) for v in reduced_vars)
         assert isinstance(terms, tuple) and len(terms) > 0
 
         assert not (isinstance(red_op, NullOp) and isinstance(bin_op, NullOp))
@@ -62,17 +63,17 @@ class Contraction(Funsor):
             assert reduced_vars and len(terms) > 1
             assert (red_op, bin_op) in DISTRIBUTIVE_OPS
 
+        fresh = frozenset()
+        bound = {v.name: v.output for v in reduced_vars}
         inputs = OrderedDict()
         for v in terms:
-            inputs.update((k, d) for k, d in v.inputs.items() if k not in reduced_vars)
+            inputs.update((k, d) for k, d in v.inputs.items() if k not in bound)
 
         if bin_op is nullop:
             output = terms[0].output
         else:
             output = reduce(lambda lhs, rhs: find_domain(bin_op, lhs, rhs),
                             [v.output for v in reversed(terms)])
-        fresh = frozenset()
-        bound = reduced_vars
         super(Contraction, self).__init__(inputs, output, fresh, bound)
         self.red_op = red_op
         self.bin_op = bin_op
@@ -127,6 +128,14 @@ class Contraction(Funsor):
                     terms.append(-gaussian.log_normalizer)
                     terms.append(term.unscaled_sample(greedy_vars, sample_inputs, rng_keys[0]))
                     result = Contraction(self.red_op, self.bin_op, self.reduced_vars, *terms)
+                elif any(isinstance(term, funsor.distribution.Distribution)
+                         and not greedy_vars.isdisjoint(term.value.inputs) for term in greedy_terms):
+                    sampled_terms = [
+                        term.unscaled_sample(greedy_vars.intersection(term.value.inputs), sample_inputs)
+                        for term in greedy_terms if isinstance(term, funsor.distribution.Distribution)
+                        and not greedy_vars.isdisjoint(term.value.inputs)
+                    ]
+                    result = Contraction(self.red_op, self.bin_op, self.reduced_vars, *(terms + sampled_terms))
                 else:
                     raise NotImplementedError('Unhandled case: {}'.format(
                         ', '.join(str(type(t)) for t in greedy_terms)))
@@ -144,16 +153,14 @@ class Contraction(Funsor):
         return result
 
     def _alpha_convert(self, alpha_subs):
-        reduced_vars = frozenset(alpha_subs.get(k, k) for k in self.reduced_vars)
-        bound_types = {}
-        for term in self.terms:
-            bound_types.update({k: term.inputs[k] for k in self.bound.intersection(term.inputs)})
-        alpha_subs = {k: to_funsor(v, bound_types[k]) for k, v in alpha_subs.items()}
+        reduced_vars = frozenset(to_funsor(alpha_subs.get(var.name, var), var.output)
+                                 for var in self.reduced_vars)
+        alpha_subs = {k: to_funsor(v, self.bound[k]) for k, v in alpha_subs.items()}
         red_op, bin_op, _, terms = super()._alpha_convert(alpha_subs)
         return red_op, bin_op, reduced_vars, terms
 
 
-GaussianMixture = Contraction[Union[ops.LogAddExpOp, NullOp], ops.AddOp, frozenset,
+GaussianMixture = Contraction[Union[ops.LogaddexpOp, NullOp], ops.AddOp, frozenset,
                               Tuple[Union[Tensor, Number], Gaussian]]
 
 
@@ -182,30 +189,35 @@ def eager_contraction_generic_to_tuple(red_op, bin_op, reduced_vars, *terms):
 
 @eager.register(Contraction, AssociativeOp, AssociativeOp, frozenset, tuple)
 def eager_contraction_generic_recursive(red_op, bin_op, reduced_vars, terms):
-    # push down leaf reductions
-    terms, reduced_vars, leaf_reduced = list(terms), frozenset(reduced_vars), False
-    for i, v in enumerate(terms):
-        unique_vars = reduced_vars.intersection(v.inputs) - \
-            frozenset().union(*(reduced_vars.intersection(vv.inputs) for vv in terms if vv is not v))
-        if unique_vars:
-            result = v.reduce(red_op, unique_vars)
-            if result is not normalize(Contraction, red_op, nullop, unique_vars, (v,)):
-                terms[i] = result
-                reduced_vars -= unique_vars
-                leaf_reduced = True
+    # Count the number of terms in which each variable is reduced.
+    counts = Counter()
+    for term in terms:
+        counts.update(reduced_vars & term.input_vars)
 
+    # push down leaf reductions
+    terms = list(terms)
+    leaf_reduced = False
+    reduced_once = frozenset(v for v, count in counts.items() if count == 1)
+    if reduced_once:
+        for i, term in enumerate(terms):
+            unique_vars = reduced_once & term.input_vars
+            if unique_vars:
+                result = term.reduce(red_op, unique_vars)
+                if result is not normalize(Contraction, red_op, nullop, unique_vars, (term,)):
+                    terms[i] = result
+                    reduced_vars -= unique_vars
+                    leaf_reduced = True
     if leaf_reduced:
         return Contraction(red_op, bin_op, reduced_vars, *terms)
 
     # exploit associativity to recursively evaluate this contraction
     # a bit expensive, but handles interpreter-imposed directionality constraints
     terms = tuple(terms)
+    reduced_twice = frozenset(v for v, count in counts.items() if count == 2)
     for i, lhs in enumerate(terms[0:-1]):
         for j_, rhs in enumerate(terms[i+1:]):
             j = i + j_ + 1
-            unique_vars = reduced_vars.intersection(lhs.inputs, rhs.inputs) - \
-                frozenset().union(*(reduced_vars.intersection(vv.inputs)
-                                    for vv in terms[:i] + terms[i+1:j] + terms[j+1:]))
+            unique_vars = reduced_twice.intersection(lhs.input_vars, rhs.input_vars)
             result = Contraction(red_op, bin_op, unique_vars, lhs, rhs)
             if result is not normalize(Contraction, red_op, bin_op, unique_vars, (lhs, rhs)):  # did we make progress?
                 # pick the first evaluable pair
@@ -225,7 +237,7 @@ def eager_contraction_to_reduce(red_op, bin_op, reduced_vars, term):
 @eager.register(Contraction, AssociativeOp, AssociativeOp, frozenset, Funsor, Funsor)
 def eager_contraction_to_binary(red_op, bin_op, reduced_vars, lhs, rhs):
 
-    if reduced_vars - (reduced_vars.intersection(lhs.inputs, rhs.inputs)):
+    if not reduced_vars.issubset(lhs.input_vars & rhs.input_vars):
         args = red_op, bin_op, reduced_vars, (lhs, rhs)
         result = eager.dispatch(Contraction, *args)(*args)
         if result is not None:
@@ -247,7 +259,7 @@ def eager_contraction_tensor(red_op, bin_op, reduced_vars, *terms):
     return _eager_contract_tensors(reduced_vars, terms, backend=backend)
 
 
-@eager.register(Contraction, ops.LogAddExpOp, ops.AddOp, frozenset, Tensor, Tensor)
+@eager.register(Contraction, ops.LogaddexpOp, ops.AddOp, frozenset, Tensor, Tensor)
 def eager_contraction_tensor(red_op, bin_op, reduced_vars, *terms):
     if not all(term.dtype == "real" for term in terms):
         raise NotImplementedError('TODO')
@@ -277,8 +289,8 @@ def _eager_contract_tensors(reduced_vars, terms, backend):
         data = data.reshape(batch_shape + event_shape)
         operands.append(data)
 
-    for k in reduced_vars:
-        del inputs[k]
+    for var in reduced_vars:
+        inputs.pop(var.name, None)
     batch_shape = tuple(v.size for v in inputs.values())
     event_shape = broadcast_shape(*(term.shape for term in terms))
     einsum_output = ("".join(symbols[k] for k in inputs) +
@@ -294,7 +306,7 @@ def _eager_contract_tensors(reduced_vars, terms, backend):
 # TODO(https://github.com/pyro-ppl/funsor/issues/238) Use a port of
 # Pyro's gaussian_tensordot() here. Until then we must eagerly add the
 # possibly-rank-deficient terms before reducing to avoid Cholesky errors.
-@eager.register(Contraction, ops.LogAddExpOp, ops.AddOp, frozenset,
+@eager.register(Contraction, ops.LogaddexpOp, ops.AddOp, frozenset,
                 GaussianMixture, GaussianMixture)
 def eager_contraction_gaussian(red_op, bin_op, reduced_vars, x, y):
     return (x + y).reduce(red_op, reduced_vars)
@@ -443,7 +455,7 @@ def binary_subtract(op, lhs, rhs):
     return lhs + -rhs
 
 
-@normalize.register(Binary, ops.DivOp, Funsor, Funsor)
+@normalize.register(Binary, ops.TruedivOp, Funsor, Funsor)
 def binary_divide(op, lhs, rhs):
     return lhs * Unary(ops.reciprocal, rhs)
 
