@@ -21,7 +21,8 @@ from funsor.gaussian import Gaussian
 from funsor.interpreter import gensym
 from funsor.tensor import (Tensor, align_tensors, dummy_numeric_array, get_default_prototype,
                            ignore_jit_warnings, numeric_array, stack)
-from funsor.terms import Funsor, FunsorMeta, Independent, Number, Variable, eager, to_data, to_funsor
+from funsor.terms import Funsor, FunsorMeta, Independent, Number, Variable, \
+    eager, reflect, to_data, to_funsor
 from funsor.util import broadcast_shape, get_backend, getargspec, lazy_property
 
 
@@ -56,12 +57,34 @@ class DistributionMeta(FunsorMeta):
     """
     def __call__(cls, *args, **kwargs):
         kwargs.update(zip(cls._ast_fields, args))
-        value = kwargs.pop('value', 'value')
-        kwargs = OrderedDict(
-            (k, to_funsor(kwargs[k], output=cls._infer_param_domain(k, getattr(kwargs[k], "shape", ()))))
-            for k in cls._ast_fields if k != 'value')
-        value = to_funsor(value, output=cls._infer_value_domain(**{k: v.output for k, v in kwargs.items()}))
-        args = numbers_to_tensors(*(tuple(kwargs.values()) + (value,)))
+        kwargs["value"] = kwargs.get("value", "value")
+        kwargs = OrderedDict((k, kwargs[k]) for k in cls._ast_fields)  # make sure args are sorted
+
+        domains = OrderedDict()
+        for k, v in kwargs.items():
+            if k == "value":
+                continue
+
+            # compute unbroadcasted param domains
+            domain = cls._infer_param_domain(k, getattr(kwargs[k], "shape", ()))
+            # use to_funsor to infer output dimensions of e.g. tensors
+            domains[k] = domain if domain is not None else to_funsor(v).output
+
+            # broadcast individual param domains with Funsor inputs
+            # this avoids .expand-ing underlying parameter tensors
+            dtype = domains[k].dtype
+            if isinstance(v, Funsor):
+                domains[k] = Array[dtype, broadcast_shape(v.shape, domains[k].shape)]
+            elif ops.is_numeric_array(v):
+                domains[k] = Array[dtype, broadcast_shape(v.shape, domains[k].shape)]
+
+        # now use the broadcasted parameter shapes to infer the event_shape
+        domains["value"] = cls._infer_value_domain(**domains)
+
+        # finally, perform conversions to funsors
+        kwargs = OrderedDict((k, to_funsor(v, output=domains[k])) for k, v in kwargs.items())
+        args = numbers_to_tensors(*kwargs.values())
+
         return super(DistributionMeta, cls).__call__(*args)
 
 
@@ -93,80 +116,129 @@ class Distribution(Funsor, metaclass=DistributionMeta):
                                ', '.join('{}={}'.format(*kv) for kv in self.params.items()))
 
     def eager_reduce(self, op, reduced_vars):
+        assert reduced_vars.issubset(self.inputs)
         if op is ops.logaddexp and isinstance(self.value, Variable) and self.value.name in reduced_vars:
             return Number(0.)  # distributions are normalized
         return super(Distribution, self).eager_reduce(op, reduced_vars)
 
-    @classmethod
-    def eager_log_prob(cls, *params):
-        inputs, tensors = align_tensors(*params)
-        params = dict(zip(cls._ast_fields, tensors))
-        value = params.pop('value')
-        data = cls.dist_class(**params).log_prob(value)
-        return Tensor(data, inputs)
-
-    @property
-    def has_rsample(self):
-        return getattr(self.dist_class, "has_rsample", False)
+    def _get_raw_dist(self):
+        """
+        Internal method for working with underlying distribution attributes
+        """
+        value_name = [name for name, domain in self.value.inputs.items()  # TODO is this right?
+                      if domain == self.value.output][0]
+        # arbitrary name-dim mapping, since we're converting back to a funsor anyway
+        name_to_dim = {name: -dim-1 for dim, (name, domain) in enumerate(self.inputs.items())
+                       if isinstance(domain.dtype, int) and name != value_name}
+        raw_dist = to_data(self, name_to_dim=name_to_dim)
+        dim_to_name = {dim: name for name, dim in name_to_dim.items()}
+        # also return value output, dim_to_name for converting results back to funsor
+        value_output = self.inputs[value_name]
+        return raw_dist, value_name, value_output, dim_to_name
 
     @property
     def has_enumerate_support(self):
         return getattr(self.dist_class, "has_enumerate_support", False)
 
+    @classmethod
+    def eager_log_prob(cls, *params):
+        params, value = params[:-1], params[-1]
+        params = params + (Variable("value", value.output),)
+        instance = reflect(cls, *params)
+        raw_dist, value_name, value_output, dim_to_name = instance._get_raw_dist()
+        assert value.output == value_output
+        name_to_dim = {v: k for k, v in dim_to_name.items()}
+        dim_to_name.update({-1 - d - len(raw_dist.batch_shape): name
+                            for d, name in enumerate(value.inputs) if name not in name_to_dim})
+        name_to_dim.update({v: k for k, v in dim_to_name.items() if v not in name_to_dim})
+        raw_log_prob = raw_dist.log_prob(to_data(value, name_to_dim=name_to_dim))
+        log_prob = to_funsor(raw_log_prob, Real, dim_to_name=dim_to_name)
+        # this logic ensures that the inputs have the canonical order
+        # implied by align_tensors, which is assumed pervasively in tests
+        inputs = OrderedDict()
+        for x in params[:-1] + (value,):
+            inputs.update(x.inputs)
+        return log_prob.align(tuple(inputs))
+
     def unscaled_sample(self, sampled_vars, sample_inputs, rng_key=None):
-        params = OrderedDict(self.params)
-        value = params.pop("value")
-        assert all(isinstance(v, (Number, Tensor)) for v in params.values())
-        assert isinstance(value, Variable) and value.name in sampled_vars
-        inputs_, tensors = align_tensors(*params.values())
-        inputs = OrderedDict(sample_inputs.items())
-        inputs.update(inputs_)
+
+        # note this should handle transforms correctly via distribution_to_data
+        raw_dist, value_name, value_output, dim_to_name = self._get_raw_dist()
+        for d, name in zip(range(len(sample_inputs), 0, -1), sample_inputs.keys()):
+            dim_to_name[-d - len(raw_dist.batch_shape)] = name
+
+        if value_name not in sampled_vars:
+            return self
+
         sample_shape = tuple(v.size for v in sample_inputs.values())
-
-        raw_dist = self.dist_class(**dict(zip(self._ast_fields[:-1], tensors)))
-        sample_args = (sample_shape,) if rng_key is None else (rng_key, sample_shape)
-        if self.has_rsample:
-            raw_sample = raw_dist.rsample(*sample_args)
+        sample_args = (sample_shape,) if get_backend() == "torch" else (rng_key, sample_shape)
+        if raw_dist.has_rsample:
+            raw_value = raw_dist.rsample(*sample_args)
         else:
-            raw_sample = ops.detach(raw_dist.sample(*sample_args))
+            raw_value = ops.detach(raw_dist.sample(*sample_args))
 
-        result = funsor.delta.Delta(value.name, Tensor(raw_sample, inputs, value.output.dtype))
-        if not self.has_rsample:
+        funsor_value = to_funsor(raw_value, output=value_output, dim_to_name=dim_to_name)
+        funsor_value = funsor_value.align(
+            tuple(sample_inputs) + tuple(inp for inp in self.inputs if inp in funsor_value.inputs))
+        result = funsor.delta.Delta(value_name, funsor_value)
+        if not raw_dist.has_rsample:
             # scaling of dice_factor by num samples should already be handled by Funsor.sample
-            raw_log_prob = raw_dist.log_prob(raw_sample)
-            dice_factor = Tensor(raw_log_prob - ops.detach(raw_log_prob), inputs)
+            raw_log_prob = raw_dist.log_prob(raw_value)
+            dice_factor = to_funsor(raw_log_prob - ops.detach(raw_log_prob),
+                                    output=self.output, dim_to_name=dim_to_name)
             result = result + dice_factor
         return result
 
     def enumerate_support(self, expand=False):
-        if not self.has_enumerate_support or not isinstance(self.value, Variable):
-            raise ValueError("cannot enumerate support of {}".format(repr(self)))
-        # arbitrary name-dim mapping, since we're converting back to a funsor anyway
-        name_to_dim = {name: -dim-1 for dim, (name, domain) in enumerate(self.inputs.items())
-                       if isinstance(domain.dtype, int) and name != self.value.name}
-        raw_dist = to_data(self, name_to_dim=name_to_dim)
+        assert self.has_enumerate_support and isinstance(self.value, Variable)
+        raw_dist, value_name, value_output, dim_to_name = self._get_raw_dist()
         raw_value = raw_dist.enumerate_support(expand=expand)
-        dim_to_name = {dim: name for name, dim in name_to_dim.items()}
-        dim_to_name[min(dim_to_name.keys(), default=0)-1] = self.value.name
-        return to_funsor(raw_value, output=self.value.output, dim_to_name=dim_to_name)
+        dim_to_name[min(dim_to_name.keys(), default=0)-1] = value_name
+        return to_funsor(raw_value, output=value_output, dim_to_name=dim_to_name)
+
+    def entropy(self):
+        raw_dist, value_name, value_output, dim_to_name = self._get_raw_dist()
+        raw_value = raw_dist.entropy()
+        return to_funsor(raw_value, output=self.output, dim_to_name=dim_to_name)
+
+    def mean(self):
+        raw_dist, value_name, value_output, dim_to_name = self._get_raw_dist()
+        raw_value = raw_dist.mean
+        return to_funsor(raw_value, output=value_output, dim_to_name=dim_to_name)
+
+    def variance(self):
+        raw_dist, value_name, value_output, dim_to_name = self._get_raw_dist()
+        raw_value = raw_dist.variance
+        return to_funsor(raw_value, output=value_output, dim_to_name=dim_to_name)
 
     def __getattribute__(self, attr):
         if attr in type(self)._ast_fields and attr != 'name':
             return self.params[attr]
         return super().__getattribute__(attr)
 
+    @classmethod
+    def _infer_value_dtype(cls, domains):
+        try:
+            support = cls.dist_class.support
+        except NotImplementedError:
+            raise NotImplementedError(f"Failed to infer dtype of {cls.dist_class.__name__}")
+        while type(support).__name__ == "IndependentConstraint":
+            support = support.base_constraint
+        if type(support).__name__ == "_IntegerInterval":
+            return int(support.upper_bound + 1)
+        return "real"
+
     @classmethod
     @functools.lru_cache(maxsize=5000)
-    def _infer_value_domain(cls, **kwargs):
-        # rely on the underlying distribution's logic to infer the event_shape given param domains
-        instance = cls.dist_class(**{k: dummy_numeric_array(domain) for k, domain in kwargs.items()},
-                                  validate_args=False)
-        out_shape = instance.event_shape
-        if type(instance.support).__name__ == "_IntegerInterval":
-            out_dtype = int(instance.support.upper_bound + 1)
-        else:
-            out_dtype = 'real'
-        return Array[out_dtype, out_shape]
+    def _infer_value_domain(cls, **domains):
+        dtype = cls._infer_value_dtype(domains)
+        # TODO implement .infer_shapes() methods on each distribution
+        # TODO fix distribution constraints by wrapping in _Independent
+        batch_shape, event_shape = infer_shapes(cls.dist_class, domains)
+        shape = batch_shape + event_shape
+        if "value" in domains:
+            shape = broadcast_shape(shape, domains["value"].shape)
+        return Array[dtype, shape]
 
     @classmethod
     @functools.lru_cache(maxsize=5000)
@@ -176,35 +248,58 @@ class Distribution(Funsor, metaclass=DistributionMeta):
         # define backend-specific distributions and overide these `infer_value_domain`,
         # `infer_param_domain` methods.
         # Because NumPyro and Pyro have the same pattern, we use name check for simplicity.
-        support_name = type(support).__name__
-        if support_name == "_Simplex":
-            output = Reals[raw_shape[-1]]
-        elif support_name == "_RealVector":
-            output = Reals[raw_shape[-1]]
-        elif support_name in ["_LowerCholesky", "_PositiveDefinite"]:
-            output = Reals[raw_shape[-2:]]
+        support_name = type(support).__name__.lstrip("_")
+
+        event_dim = 0
+        while support_name == "IndependentConstraint":
+            event_dim += support.reinterpreted_batch_ndims
+            support = support.base_constraint
+            support_name = type(support).__name__.lstrip("_")
+
+        if support_name == "Simplex":
+            output = Reals[raw_shape[-1 - event_dim:]]
+        elif support_name == "RealVector":
+            output = Reals[raw_shape[-1 - event_dim:]]
+        elif support_name in ["LowerCholesky", "PositiveDefinite"]:
+            output = Reals[raw_shape[-2 - event_dim:]]
         # resolve the issue: logits's constraints are real (instead of real_vector)
         # for discrete multivariate distributions in Pyro
-        elif support_name == "_Real":
+        elif support_name == "Real":
             if name == "logits" and (
                     "probs" in cls.dist_class.arg_constraints
-                    and type(cls.dist_class.arg_constraints["probs"]).__name__ == "_Simplex"):
-                output = Reals[raw_shape[-1]]
+                    and type(cls.dist_class.arg_constraints["probs"]).__name__.lstrip("_") == "Simplex"):
+                output = Reals[raw_shape[-1 - event_dim:]]
             else:
-                output = Real
-        elif support_name in ("_Interval", "_GreaterThan", "_LessThan"):
-            output = Real
+                output = Reals[raw_shape[len(raw_shape) - event_dim:]]
+        elif support_name in ("Interval", "GreaterThan", "LessThan"):
+            output = Reals[raw_shape[len(raw_shape) - event_dim:]]
         else:
             output = None
         return output
 
 
+def infer_shapes(dist_class, domains):
+    arg_shapes = {k: domain.shape for k, domain in domains.items() if k != "value"}
+    try:
+        return dist_class.infer_shapes(**arg_shapes)
+    except (AttributeError, NotImplementedError):
+        pass
+        # warnings.warn(f"Failed to infer shape for {dist_class.__name__}, "
+        #               "falling back to expensive instance construction")
+
+    # Rely on the underlying distribution's logic to infer the event_shape
+    # given param domains.
+    args = {k: dummy_numeric_array(domain)
+            for k, domain in domains.items() if k != "value"}
+    instance = dist_class(**args, validate_args=False)
+    return instance.batch_shape, instance.event_shape
+
+
 ################################################################################
 # Distribution Wrappers
 ################################################################################
 
-
-def make_dist(backend_dist_class, param_names=()):
+def make_dist(backend_dist_class, param_names=(), generate_eager=True, generate_to_funsor=True):
     if not param_names:
         param_names = tuple(name for name in inspect.getfullargspec(backend_dist_class.__init__)[0][1:]
                             if name in backend_dist_class.arg_constraints)
@@ -218,13 +313,19 @@ def make_dist(backend_dist_class, param_names=()):
         '__init__': dist_init,
     })
 
-    eager.register(dist_class, *((Tensor,) * (len(param_names) + 1)))(dist_class.eager_log_prob)
+    if generate_eager:
+        eager.register(dist_class, *((Tensor,) * (len(param_names) + 1)))(dist_class.eager_log_prob)
+
+    if generate_to_funsor:
+        to_funsor.register(backend_dist_class)(functools.partial(backenddist_to_funsor, dist_class))
 
     return dist_class
 
 
 FUNSOR_DIST_NAMES = [
     ('Beta', ('concentration1', 'concentration0')),
+    ("Cauchy", ()),
+    ("Chi2", ()),
     ('BernoulliProbs', ('probs',)),
     ('BernoulliLogits', ('logits',)),
     ('Binomial', ('total_count', 'probs')),
@@ -233,8 +334,15 @@ FUNSOR_DIST_NAMES = [
     ('Delta', ('v', 'log_density')),
     ('Dirichlet', ('concentration',)),
     ('DirichletMultinomial', ('concentration', 'total_count')),
+    ("Exponential", ()),
     ('Gamma', ('concentration', 'rate')),
     ('GammaPoisson', ('concentration', 'rate')),
+    ("Geometric", ("probs",)),
+    ("Gumbel", ()),
+    ("HalfCauchy", ()),
+    ("HalfNormal", ()),
+    ("Laplace", ()),
+    ("LowRankMultivariateNormal", ()),
     ('Multinomial', ('total_count', 'probs')),
     ('MultivariateNormal', ('loc', 'scale_tril')),
     ('NonreparameterizedBeta', ('concentration1', 'concentration0')),
@@ -242,8 +350,11 @@ FUNSOR_DIST_NAMES = [
     ('NonreparameterizedGamma', ('concentration', 'rate')),
     ('NonreparameterizedNormal', ('loc', 'scale')),
     ('Normal', ('loc', 'scale')),
-    ('Poisson', ('rate',)),
-    ('VonMises', ('loc', 'concentration')),
+    ("Pareto", ()),
+    ("Poisson", ()),
+    ("StudentT", ()),
+    ("Uniform", ()),
+    ("VonMises", ()),
 ]
 
 
@@ -251,9 +362,7 @@ FUNSOR_DIST_NAMES = [
 # Converting backend Distributions to funsors
 ###############################################
 
-def backenddist_to_funsor(backend_dist, output=None, dim_to_name=None):
-    funsor_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
-    funsor_dist_class = getattr(funsor_dist, type(backend_dist).__name__.split("Wrapper_")[-1])
+def backenddist_to_funsor(funsor_dist_class, backend_dist, output=None, dim_to_name=None):
     params = [to_funsor(
             getattr(backend_dist, param_name),
             output=funsor_dist_class._infer_param_domain(
@@ -275,24 +384,44 @@ def indepdist_to_funsor(backend_dist, output=None, dim_to_name=None):
     return result
 
 
+def expandeddist_to_funsor(backend_dist, output=None, dim_to_name=None):
+
+    funsor_base_dist = to_funsor(backend_dist.base_dist, output=output, dim_to_name=dim_to_name)
+    if not dim_to_name:
+        assert not backend_dist.batch_shape
+        return funsor_base_dist
+
+    name_to_dim = {name: dim for dim, name in dim_to_name.items()}
+    raw_expanded_params = {}
+    for name, funsor_param in funsor_base_dist.params.items():
+        if name == "value":
+            continue
+        raw_param = to_data(funsor_param, name_to_dim=name_to_dim)
+        raw_expanded_params[name] = ops.expand(raw_param, backend_dist.batch_shape + funsor_param.shape)
+
+    raw_expanded_dist = type(backend_dist.base_dist)(**raw_expanded_params)
+    return to_funsor(raw_expanded_dist, output, dim_to_name)
+
+
 def maskeddist_to_funsor(backend_dist, output=None, dim_to_name=None):
     mask = to_funsor(ops.astype(backend_dist._mask, 'float32'), output=output, dim_to_name=dim_to_name)
     funsor_base_dist = to_funsor(backend_dist.base_dist, output=output, dim_to_name=dim_to_name)
     return mask * funsor_base_dist
 
 
+# converts TransformedDistributions
 def transformeddist_to_funsor(backend_dist, output=None, dim_to_name=None):
-    raise NotImplementedError("TODO implement conversion of TransformedDistribution")
-
-
-def mvndist_to_funsor(backend_dist, output=None, dim_to_name=None, real_inputs=OrderedDict()):
-    funsor_dist = backenddist_to_funsor(backend_dist, output=output, dim_to_name=dim_to_name)
-    if len(real_inputs) == 0:
-        return funsor_dist
-    discrete, gaussian = funsor_dist(value="value").terms
-    inputs = OrderedDict((k, v) for k, v in gaussian.inputs.items() if v.dtype != 'real')
-    inputs.update(real_inputs)
-    return discrete + Gaussian(gaussian.info_vec, gaussian.precision, inputs)
+    dist_module = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()]).dist
+    base_dist, transforms = backend_dist, []
+    while isinstance(base_dist, dist_module.TransformedDistribution):
+        transforms = base_dist.transforms + transforms
+        base_dist = base_dist.base_dist
+    funsor_base_dist = to_funsor(base_dist, output=output, dim_to_name=dim_to_name)
+    # TODO make this work with transforms that change the output type
+    transform = to_funsor(dist_module.transforms.ComposeTransform(transforms),
+                          funsor_base_dist.inputs["value"], dim_to_name)
+    _, inv_transform, ldj = funsor.delta.solve(transform, to_funsor("value", funsor_base_dist.inputs["value"]))
+    return -ldj + funsor_base_dist(value=inv_transform)
 
 
 class CoerceDistributionToFunsor:
@@ -370,12 +499,38 @@ class CoerceDistributionToFunsor:
 
 @to_data.register(Distribution)
 def distribution_to_data(funsor_dist, name_to_dim=None):
-    pyro_dist_class = funsor_dist.dist_class
-    params = [to_data(getattr(funsor_dist, param_name), name_to_dim=name_to_dim)
-              for param_name in funsor_dist._ast_fields if param_name != 'value']
-    pyro_dist = pyro_dist_class(**dict(zip(funsor_dist._ast_fields[:-1], params)))
     funsor_event_shape = funsor_dist.value.output.shape
+
+    # attempt to generically infer the independent output dimensions
+    domains = {k: v.output for k, v in funsor_dist.params.items()}
+    indep_shape, _ = infer_shapes(funsor_dist.dist_class, domains)
+
+    params = []
+    for param_name, funsor_param in zip(funsor_dist._ast_fields, funsor_dist._ast_values[:-1]):
+        param = to_data(funsor_param, name_to_dim=name_to_dim)
+
+        # infer the independent dimensions of each parameter separately, since we chose to keep them unbroadcasted
+        param_event_shape = getattr(funsor_dist._infer_param_domain(param_name, funsor_param.output.shape), "shape", ())
+        param_indep_shape = funsor_param.output.shape[:len(funsor_param.output.shape) - len(param_event_shape)]
+        for i in range(max(0, len(indep_shape) - len(param_indep_shape))):
+            # add singleton event dimensions, leave broadcasting/expanding to backend
+            param = ops.unsqueeze(param, -1 - len(funsor_param.output.shape))
+
+        params.append(param)
+
+    pyro_dist = funsor_dist.dist_class(**dict(zip(funsor_dist._ast_fields[:-1], params)))
     pyro_dist = pyro_dist.to_event(max(len(funsor_event_shape) - len(pyro_dist.event_shape), 0))
+
+    # TODO get this working for all backends
+    if not isinstance(funsor_dist.value, Variable):
+        if get_backend() != "torch":
+            raise NotImplementedError("transformed distributions not yet supported under this backend,"
+                                      "try set_backend('torch')")
+        inv_value = funsor.delta.solve(funsor_dist.value, Variable("value", funsor_dist.value.output))[1]
+        transforms = to_data(inv_value, name_to_dim=name_to_dim)
+        backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()]).dist
+        pyro_dist = backend_dist.TransformedDistribution(pyro_dist, transforms)
+
     if pyro_dist.event_shape != funsor_event_shape:
         raise ValueError("Event shapes don't match, something went wrong")
     return pyro_dist
@@ -383,7 +538,20 @@ def distribution_to_data(funsor_dist, name_to_dim=None):
 
 @to_data.register(Independent[typing.Union[Independent, Distribution], str, str, str])
 def indep_to_data(funsor_dist, name_to_dim=None):
-    raise NotImplementedError("TODO implement conversion of Independent")
+    if not isinstance(funsor_dist.fn, (Independent, Distribution, Gaussian)):
+        raise NotImplementedError(f"cannot convert {funsor_dist} to data")
+    name_to_dim = OrderedDict((name, dim - 1) for name, dim in name_to_dim.items())
+    name_to_dim.update({funsor_dist.bint_var: -1})
+    backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()]).dist
+    result = to_data(funsor_dist.fn, name_to_dim=name_to_dim)
+
+    # collapse nested Independents into a single Independent for conversion
+    reinterpreted_batch_ndims = 1
+    while isinstance(result, backend_dist.Independent):
+        result = result.base_dist
+        reinterpreted_batch_ndims += 1
+
+    return backend_dist.Independent(result, reinterpreted_batch_ndims)
 
 
 @to_data.register(Gaussian)
@@ -556,8 +724,20 @@ def eager_beta_bernoulli(red_op, bin_op, reduced_vars, x, y):
                                        backend_dist.Binomial(total_count=1, probs=y.probs, value=y.value))
 
 
+def eager_dirichlet_categorical(red_op, bin_op, reduced_vars, x, y):
+    dirichlet_reduction = x.input_vars & reduced_vars
+    if dirichlet_reduction:
+        backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+        identity = Tensor(ops.new_eye(funsor.tensor.get_default_prototype(), x.concentration.shape))
+        return backend_dist.DirichletMultinomial(concentration=x.concentration,
+                                                 total_count=1,
+                                                 value=identity[y.value])
+    else:
+        return eager(Contraction, red_op, bin_op, reduced_vars, (x, y))
+
+
 def eager_dirichlet_multinomial(red_op, bin_op, reduced_vars, x, y):
-    dirichlet_reduction = frozenset(x.inputs).intersection(reduced_vars)
+    dirichlet_reduction = x.input_vars & reduced_vars
     if dirichlet_reduction:
         backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
         return backend_dist.DirichletMultinomial(concentration=x.concentration,
@@ -567,8 +747,41 @@ def eager_dirichlet_multinomial(red_op, bin_op, reduced_vars, x, y):
         return eager(Contraction, red_op, bin_op, reduced_vars, (x, y))
 
 
+def eager_plate_multinomial(op, x, reduced_vars):
+    if not reduced_vars.isdisjoint(x.probs.input_vars):
+        return None
+    if not reduced_vars.issubset(x.value.input_vars):
+        return None
+
+    backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
+    total_count = x.total_count
+    for v in reduced_vars:
+        if v.name in total_count.inputs:
+            total_count = total_count.reduce(ops.add, v)
+        else:
+            total_count = total_count * v.output.size
+    return backend_dist.Multinomial(total_count=total_count,
+                                    probs=x.probs,
+                                    value=x.value.reduce(ops.add, reduced_vars))
+
+
+def _log_beta(x, y):
+    return ops.lgamma(x) + ops.lgamma(y) - ops.lgamma(x + y)
+
+
+def eager_gamma_gamma(red_op, bin_op, reduced_vars, x, y):
+    gamma_reduction = x.input_vars & reduced_vars
+    if gamma_reduction:
+        unnormalized = (y.concentration - 1) * ops.log(y.value) \
+            - (y.concentration + x.concentration) * ops.log(y.value + x.rate)
+        const = -x.concentration * ops.log(x.rate) + _log_beta(y.concentration, x.concentration)
+        return unnormalized - const
+    else:
+        return eager(Contraction, red_op, bin_op, reduced_vars, (x, y))
+
+
 def eager_gamma_poisson(red_op, bin_op, reduced_vars, x, y):
-    gamma_reduction = frozenset(x.inputs).intersection(reduced_vars)
+    gamma_reduction = x.input_vars & reduced_vars
     if gamma_reduction:
         backend_dist = import_module(BACKEND_TO_DISTRIBUTIONS_BACKEND[get_backend()])
         return backend_dist.GammaPoisson(concentration=x.concentration,
