@@ -1,31 +1,37 @@
 # Copyright Contributors to the Pyro project.
 # SPDX-License-Identifier: Apache-2.0
 
+import numbers
+
 import jax.numpy as np
 import numpy as onp
 from jax import lax
 from jax.core import Tracer
 from jax.interpreters.xla import DeviceArray
 from jax.scipy.linalg import cho_solve, solve_triangular
-from jax.scipy.special import expit, logsumexp
+from jax.scipy.special import expit, gammaln, logsumexp
 
 import funsor.ops as ops
 
-
 ################################################################################
 # Register Ops
 ################################################################################
 
 array = (onp.generic, onp.ndarray, DeviceArray, Tracer)
-ops.clamp.register(array, object, object)(np.clip)
+ops.atanh.register(array)(np.arctanh)
+ops.clamp.register(array, numbers.Number, numbers.Number)(np.clip)
+ops.clamp.register(array, numbers.Number, type(None))(np.clip)
+ops.clamp.register(array, type(None), numbers.Number)(np.clip)
+ops.clamp.register(array, type(None), type(None))(np.clip)
 ops.exp.register(array)(np.exp)
-ops.full_like.register(array, object)(np.full_like)
+ops.full_like.register(array, numbers.Number)(np.full_like)
 ops.log1p.register(array)(np.log1p)
 ops.max.register(array)(np.maximum)
 ops.min.register(array)(np.minimum)
 ops.permute.register(array, (tuple, list))(np.transpose)
 ops.sigmoid.register(array)(expit)
 ops.sqrt.register(array)(np.sqrt)
+ops.tanh.register(array)(np.tanh)
 ops.transpose.register(array, int, int)(np.swapaxes)
 ops.unsqueeze.register(array, int)(np.expand_dims)
 
@@ -119,6 +125,16 @@ def _is_numeric_array(x):
     return True
 
 
+@ops.isnan.register(array)
+def _isnan(x):
+    return np.isnan(x)
+
+
+@ops.lgamma.register(array)
+def _lgamma(x):
+    return gammaln(x)
+
+
 @ops.log.register(array)
 def _log(x):
     return np.log(x)
@@ -149,7 +165,7 @@ def _min(x, y):
     return np.minimum(x, y)
 
 
-# TODO: replace (int, float) by object
+# TODO: replace (int, float) by numbers.Number
 @ops.min.register((int, float), array)
 def _min(x, y):
     return np.clip(y, a_min=None, a_max=x)
