@@ -49,7 +49,7 @@ class Distribution(object):
 
     # Draw a sample.
     def __call__(self):
-        with funsor.interpreter.interpretation(funsor.terms.eager):
+        with funsor.interpretation(funsor.terms.eager):
             dist = self.funsor_dist(value='value')
             delta = dist.sample(frozenset(['value']), sample_inputs=self.sample_inputs)
         if isinstance(delta, funsor.cnf.Contraction):
@@ -62,9 +62,9 @@ class Distribution(object):
     # Similar to torch.distributions.Distribution.expand().
     def expand_inputs(self, name, size):
         if name in self.funsor_dist.inputs:
-            assert self.funsor_dist.inputs[name] == funsor.bint(int(size))
+            assert self.funsor_dist.inputs[name] == funsor.Bint[int(size)]
             return self
-        inputs = OrderedDict([(name, funsor.bint(int(size)))])
+        inputs = OrderedDict([(name, funsor.Bint[int(size)])])
         if self.sample_inputs:
             inputs.update(self.sample_inputs)
         return Distribution(self.funsor_dist, sample_inputs=inputs)
@@ -211,7 +211,7 @@ def tensor_to_funsor(value, cond_indep_stack, output):
         else:
             frame = cond_indep_stack[dim - len(batch_shape)]
             assert size == frame.size, (size, frame)
-            inputs[frame.name] = funsor.bint(int(size))
+            inputs[frame.name] = funsor.Bint[int(size)]
     value = funsor.tensor.Tensor(data, inputs, output.dtype)
     assert value.output == output
     return value
@@ -298,7 +298,7 @@ def param(name, init_value=None, constraint=torch.distributions.constraints.real
     if init_value is not None:
         if event_dim is None:
             event_dim = init_value.dim()
-        output = funsor.reals(*init_value.shape[init_value.dim() - event_dim:])
+        output = funsor.Reals[init_value.shape[init_value.dim() - event_dim:]]
 
     def fn(init_value, constraint):
         if name in PARAM_STORE:
@@ -508,7 +508,7 @@ class ELBO(object):
 # This is a wrapper for compatibility with full Pyro.
 class Trace_ELBO(ELBO):
     def __call__(self, model, guide, *args, **kwargs):
-        with funsor.montecarlo.monte_carlo_interpretation():
+        with funsor.interpretation(funsor.montecarlo.MonteCarlo()):
             return elbo(model, guide, *args, **kwargs)
 
 
@@ -521,7 +521,7 @@ class TraceEnum_ELBO(ELBO):
     # TODO allow mixing of sampling and exact integration
     def __call__(self, model, guide, *args, **kwargs):
         if self.options.get("optimize", None):
-            with funsor.interpreter.interpretation(funsor.optimizer.optimize):
+            with funsor.interpretation(funsor.optimizer.optimize):
                 elbo_expr = elbo(model, guide, *args, **kwargs)
             return funsor.reinterpret(elbo_expr)
         return elbo(model, guide, *args, **kwargs)
@@ -561,7 +561,7 @@ class Jit(object):
                     self._param_trace[name]["value"] = constrained_param
                 result = replay(self.fn, guide_trace=self._param_trace)(*args)
                 assert not result.inputs
-                assert result.output == funsor.reals()
+                assert result.output == funsor.Real
                 return funsor.to_data(result)
 
             with validation_enabled(False), warnings.catch_warnings():
