@@ -45,6 +45,127 @@ def _partition(terms, sum_vars):
     return components
 
 
+def _unroll_plate(factors, var_to_ordinal, sum_vars, plate, step):
+    # size of the plate
+    size = next(iter(f.inputs[plate].size for f in factors if plate in f.inputs))
+    # history of the plate
+    history = 1 if step else 0
+
+    # replicated variables
+    plate_vars = set()
+    for var, ordinal in var_to_ordinal.items():
+        if plate in ordinal:
+            plate_vars.add(var)
+
+    # make sure that all vars in the plate are being unrolled
+    assert plate_vars.issubset(sum_vars)
+
+    # unroll variables
+    for var in plate_vars:
+        sum_vars -= frozenset({var})
+        if var in step.keys():
+            new_var = frozenset({"{}_{}".format(var.split("_")[0], i)
+                                 for i in range(size)})
+        elif var in step.values():
+            new_var = frozenset({"{}_{}".format(var.split("_")[0], i+history)
+                                 for i in range(size)})
+        else:
+            new_var = frozenset({"{}_{}".format(var, i+history)
+                                 for i in range(size)})
+        sum_vars |= new_var
+        ordinal = var_to_ordinal.pop(var)
+        new_ordinal = ordinal.difference({plate})
+        var_to_ordinal.update({v: new_ordinal for v in new_var})
+
+    # unroll factors
+    unrolled_factors = []
+    for factor in factors:
+        if plate in factor.inputs:
+            f_vars = plate_vars.intersection(factor.inputs)
+            prev_to_var = {key: key.split("_")[0] for key in step.keys()}
+            curr_to_var = {value: value.split("_")[0] for value in step.values()}
+            nonmarkov_vars = f_vars - set(step.keys()) - set(step.values())
+            unrolled_factors.extend([factor(
+                    **{plate: i},
+                    **{var: "{}_{}".format(var, i+history) for var in nonmarkov_vars},
+                    **{curr: "{}_{}".format(var, i+history) for curr, var in curr_to_var.items()},
+                    **{prev: "{}_{}".format(var, i) for prev, var in prev_to_var.items()},
+                ) for i in range(size)])
+        else:
+            unrolled_factors.append(factor)
+
+    return unrolled_factors, var_to_ordinal, sum_vars
+
+
+def partial_unroll(factors, eliminate=frozenset(), plate_to_step=dict()):
+    """
+    Performs partial unrolling of plated factor graphs to standard factor graphs.
+    Only plates with history={0, 1} are supported.
+
+    For plates (history=0) unrolling operation appends ``_{i}`` suffix
+    to variable names for index ``i`` in the plate (e.g., "x"->"x_0" for i=0).
+    For markov dimensions (history=1) unrolling operation renames the suffixes
+    ``var_prev`` to ``var_{i}`` and ``var_curr`` to ``var_{i+1}`` for index ``i``
+    (e.g., "x_prev"->"x_0" and "x_curr"->"x_1" for i=0).
+    Markov vars are assumed to have names that follow ``var_suffix`` formatting
+    and specifically ``var_0`` for the initial factor (e.g.,
+    ``("x_0", "x_prev", "x_curr")`` for history=1).
+
+    :param factors: A collection of funsors.
+    :type factors: tuple or list
+    :param frozenset eliminate: A set of free variables to unroll,
+        including both sum variables and product variable.
+    :param dict plate_to_step: A dict mapping markov dimensions to
+        ``step`` collections that contain ordered sequences of Markov variable names
+        (e.g., ``{"time": frozenset({("x_0", "x_prev", "x_curr")})}``).
+        Plates are passed with an empty ``step``.
+    :return: a list of partially unrolled Funsors,
+        a frozenset of partially unrolled variable names,
+        and a frozenset of remaining plates.
+    """
+    assert isinstance(factors, (tuple, list))
+    assert all(isinstance(f, Funsor) for f in factors)
+    assert isinstance(eliminate, frozenset)
+    assert isinstance(plate_to_step, dict)
+    assert all(len(set(var.split("_")[0] for var in chain)) == 1
+               and chain[0].endswith("_0")
+               for step in plate_to_step.values() if step
+               for chain in step)
+    # process plate_to_step
+    plate_to_step = plate_to_step.copy()
+    for key, step in plate_to_step.items():
+        # make a dict step e.g. {"x_prev": "x_curr"}; specific to history = 1
+        plate_to_step[key] = {s[1]: s[2] for s in step}
+
+    plates = frozenset(plate_to_step.keys())
+    sum_vars = eliminate - plates
+    unrolled_plates = {k: v for (k, v) in plate_to_step.items() if k in eliminate}
+    remaining_plates = {k: v for (k, v) in plate_to_step.items() if k not in eliminate}
+
+    var_to_ordinal = {}
+    for f in factors:
+        ordinal = plates.intersection(f.inputs)
+        for var in set(f.inputs) - plates:
+            var_to_ordinal[var] = var_to_ordinal.get(var, ordinal) & ordinal
+
+    # first unroll plates with history=1 and highest ordinal
+    # then unroll plates with history=0
+    plate_to_order = {}
+    for plate, step in unrolled_plates.items():
+        if step:
+            plate_to_order[plate] = max(len(var_to_ordinal[s]) for s in step)
+        else:
+            plate_to_order[plate] = 0
+
+    # unroll one plate at a time
+    for plate in sorted(unrolled_plates.keys(), key=lambda p: plate_to_order[p], reverse=True):
+        step = unrolled_plates[plate]
+        factors, var_to_ordinal, sum_vars = \
+            _unroll_plate(factors, var_to_ordinal, sum_vars, plate, step)
+
+    return factors, sum_vars, remaining_plates
+
+
 def partial_sum_product(sum_op, prod_op, factors, eliminate=frozenset(), plates=frozenset()):
     """
     Performs partial sum-product contraction of a collection of factors.
@@ -93,6 +214,115 @@ def partial_sum_product(sum_op, prod_op, factors, eliminate=frozenset(), plates=
     return results
 
 
+def modified_partial_sum_product(sum_op, prod_op, factors,
+                                 eliminate=frozenset(), plate_to_step=dict()):
+    """
+    Generalization of the tensor variable elimination algorithm of
+    :func:`funsor.sum_product.partial_sum_product` to handle markov dimensions
+    in addition to plate dimensions. Markov dimensions in transition factors
+    are eliminated efficiently using the parallel-scan algorithm in
+    :func:`funsor.sum_product.sequential_sum_product`. The resulting factors are then
+    combined with the initial factors and final states are eliminated. Therefore,
+    when Markov dimension is eliminated ``factors`` has to contain a pairs of
+    initial factors and transition factors.
+
+    :param ~funsor.ops.AssociativeOp sum_op: A semiring sum operation.
+    :param ~funsor.ops.AssociativeOp prod_op: A semiring product operation.
+    :param factors: A collection of funsors.
+    :type factors: tuple or list
+    :param frozenset eliminate: A set of free variables to eliminate,
+        including both sum variables and product variable.
+    :param dict plate_to_step: A dict mapping markov dimensions to
+        ``step`` collections that contain ordered sequences of Markov variable names
+        (e.g., ``{"time": frozenset({("x_0", "x_prev", "x_curr")})}``).
+        Plates are passed with an empty ``step``.
+    :return: a list of partially contracted Funsors.
+    :rtype: list
+    """
+    assert callable(sum_op)
+    assert callable(prod_op)
+    assert isinstance(factors, (tuple, list))
+    assert all(isinstance(f, Funsor) for f in factors)
+    assert isinstance(eliminate, frozenset)
+    assert isinstance(plate_to_step, dict)
+    # process plate_to_step
+    plate_to_step = plate_to_step.copy()
+    prev_to_init = {}
+    for key, step in plate_to_step.items():
+        # map prev to init; works for any history > 0
+        for chain in step:
+            init, prev = chain[:len(chain)//2], chain[len(chain)//2:-1]
+            prev_to_init.update(zip(prev, init))
+        # convert step to dict type required for MarkovProduct
+        plate_to_step[key] = {chain[1]: chain[2] for chain in step}
+
+    plates = frozenset(plate_to_step.keys())
+    sum_vars = eliminate - plates
+    prod_vars = eliminate.intersection(plates)
+    markov_sum_vars = frozenset()
+    for step in plate_to_step.values():
+        markov_sum_vars |= frozenset(step.keys()) | frozenset(step.values())
+    markov_sum_vars &= sum_vars
+    markov_prod_vars = frozenset(k for k, v in plate_to_step.items() if v and k in eliminate)
+    markov_sum_to_prod = defaultdict(set)
+    for markov_prod in markov_prod_vars:
+        for k, v in plate_to_step[markov_prod].items():
+            markov_sum_to_prod[k].add(markov_prod)
+            markov_sum_to_prod[v].add(markov_prod)
+
+    var_to_ordinal = {}
+    ordinal_to_factors = defaultdict(list)
+    for f in factors:
+        ordinal = plates.intersection(f.inputs)
+        ordinal_to_factors[ordinal].append(f)
+        for var in sum_vars.intersection(f.inputs):
+            var_to_ordinal[var] = var_to_ordinal.get(var, ordinal) & ordinal
+
+    ordinal_to_vars = defaultdict(set)
+    for var, ordinal in var_to_ordinal.items():
+        ordinal_to_vars[ordinal].add(var)
+
+    results = []
+    while ordinal_to_factors:
+        leaf = max(ordinal_to_factors, key=len)
+        leaf_factors = ordinal_to_factors.pop(leaf)
+        leaf_reduce_vars = ordinal_to_vars[leaf]
+        for (group_factors, group_vars) in _partition(leaf_factors, leaf_reduce_vars | markov_prod_vars):
+            # eliminate non markov vars
+            nonmarkov_vars = group_vars - markov_sum_vars - markov_prod_vars
+            f = reduce(prod_op, group_factors).reduce(sum_op, nonmarkov_vars)
+            # eliminate markov vars
+            markov_vars = group_vars.intersection(markov_sum_vars)
+            if markov_vars:
+                markov_prod_var = [markov_sum_to_prod[var] for var in markov_vars]
+                assert all(p == markov_prod_var[0] for p in markov_prod_var)
+                if len(markov_prod_var[0]) != 1:
+                    raise ValueError("intractable!")
+                time = next(iter(markov_prod_var[0]))
+                for v in sum_vars.intersection(f.inputs):
+                    if time in var_to_ordinal[v] and var_to_ordinal[v] < leaf:
+                        raise ValueError("intractable!")
+                time_var = Variable(time, f.inputs[time])
+                group_step = {k: v for (k, v) in plate_to_step[time].items() if v in markov_vars}
+                f = MarkovProduct(sum_op, prod_op, f, time_var, group_step)
+                f = f.reduce(sum_op, frozenset(group_step.values()))
+                f = f(**prev_to_init)
+
+            remaining_sum_vars = sum_vars.intersection(f.inputs)
+
+            if not remaining_sum_vars:
+                results.append(f.reduce(prod_op, leaf & prod_vars - markov_prod_vars))
+            else:
+                new_plates = frozenset().union(
+                    *(var_to_ordinal[v] for v in remaining_sum_vars))
+                if new_plates == leaf:
+                    raise ValueError("intractable!")
+                f = f.reduce(prod_op, leaf - new_plates - markov_prod_vars)
+                ordinal_to_factors[new_plates].append(f)
+
+    return results
+
+
 def sum_product(sum_op, prod_op, factors, eliminate=frozenset(), plates=frozenset()):
     """
     Performs sum-product contraction of a collection of factors.
@@ -166,7 +396,8 @@ def sequential_sum_product(sum_op, prod_op, trans, time, step):
     drop = tuple("_drop_{}".format(i) for i in range(len(step)))
     prev_to_drop = dict(zip(step.keys(), drop))
     curr_to_drop = dict(zip(step.values(), drop))
-    drop = frozenset(drop)
+    drop = frozenset(Variable(v, trans.inputs[k])
+                     for k, v in curr_to_drop.items())
 
     time, duration = time.name, time.output.size
     while duration > 1:
@@ -243,45 +474,49 @@ def mixed_sequential_sum_product(sum_op, prod_op, trans, time, step, num_segment
     return second_stage_result
 
 
+def _get_shift(name):
+    """helper function used internally in sarkka_bilmes_product"""
+    return len(re.search(r"^(_PREV_)*", name).group(0)) // 6
+
+
+def _shift_name(name, t):
+    """helper function used internally in sarkka_bilmes_product"""
+    if t >= 0:
+        return t * "_PREV_" + name
+    return name.replace("_PREV_" * -t, "", 1)
+
+
+def _shift_funsor(f, t, global_vars):
+    """helper function used internally in sarkka_bilmes_product"""
+    if t == 0:
+        return f
+    return f(**{name: _shift_name(name, t) for name in f.inputs if name not in global_vars})
+
+
 def naive_sarkka_bilmes_product(sum_op, prod_op, trans, time_var, global_vars=frozenset()):
 
     assert isinstance(global_vars, frozenset)
 
     time = time_var.name
+    global_vars |= {time}
 
-    def get_shift(name):
-        return len(re.search("^P*", name).group(0))
-
-    def shift_name(name, t):
-        return t * "P" + name
-
-    def shift_funsor(f, t):
-        if t == 0:
-            return f
-        return f(**{name: shift_name(name, t) for name in f.inputs
-                    if name != time and name not in global_vars})
-
-    lags = {get_shift(name) for name in trans.inputs if name != time}
+    lags = {_get_shift(name) for name in trans.inputs if name != time}
     lags.discard(0)
     if not lags:
         return naive_sequential_sum_product(sum_op, prod_op, trans, time_var, {})
 
-    period = int(reduce(lambda a, b: a * b // gcd(a, b), list(lags)))
+    original_names = frozenset(name for name in trans.inputs
+                               if name not in global_vars and not name.startswith("_PREV_"))
 
     duration = trans.inputs[time].size
-    if duration % period:
-        raise NotImplementedError("TODO handle partial windows")
 
     result = trans(**{time: duration - 1})
-    original_names = frozenset(name for name in trans.inputs
-                               if name != time and name not in global_vars
-                               and not name.startswith("P"))
-    for t in range(trans.inputs[time].size - 2, -1, -1):
-        result = prod_op(shift_funsor(trans(**{time: t}), duration - t - 1), result)
-        sum_vars = frozenset(shift_name(name, duration - t - 1) for name in original_names)
+    for t in range(duration - 2, -1, -1):
+        result = prod_op(_shift_funsor(trans(**{time: t}), duration - t - 1, global_vars), result)
+        sum_vars = frozenset(_shift_name(name, duration - t - 1) for name in original_names)
         result = result.reduce(sum_op, sum_vars)
 
-    result = result(**{name: name.replace("P" * duration, "P") for name in result.inputs})
+    result = result(**{name: _shift_name(name, -duration + 1) for name in result.inputs})
     return result
 
 
@@ -290,50 +525,60 @@ def sarkka_bilmes_product(sum_op, prod_op, trans, time_var, global_vars=frozense
     assert isinstance(global_vars, frozenset)
 
     time = time_var.name
+    global_vars |= {time}
 
-    def get_shift(name):
-        return len(re.search("^P*", name).group(0))
-
-    def shift_name(name, t):
-        return t * "P" + name
-
-    def shift_funsor(f, t):
-        if t == 0:
-            return f
-        return f(**{name: shift_name(name, t) for name in f.inputs
-                    if name != time and name not in global_vars})
-
-    lags = {get_shift(name) for name in trans.inputs if name != time}
+    lags = {_get_shift(name) for name in trans.inputs if name != time}
     lags.discard(0)
     if not lags:
         return sequential_sum_product(sum_op, prod_op, trans, time_var, {})
 
     period = int(reduce(lambda a, b: a * b // gcd(a, b), list(lags)))
     original_names = frozenset(name for name in trans.inputs
-                               if name != time and name not in global_vars
-                               and not name.startswith("P"))
+                               if name not in global_vars and not name.startswith("_PREV_"))
     renamed_factors = []
     duration = trans.inputs[time].size
-    if duration % period:
-        raise NotImplementedError("TODO handle partial windows")
+    if duration % period != 0:
+        remaining_duration = duration % period
+        truncated_duration = duration - remaining_duration
+        if truncated_duration == 0:
+            result = trans(**{time: remaining_duration - 1})
+            remaining_duration -= 1
+        else:
+            # chop off the rightmost set of complete chunks from trans,
+            # then recursively call sarkka_bilmes_product on truncated factor
+            result = sarkka_bilmes_product(
+                sum_op, prod_op,
+                trans(**{time: Slice(time, remaining_duration, duration, 1, duration)}),
+                Variable(time, Bint[truncated_duration]),
+                global_vars - {time}, num_periods
+            )
+
+        # sequentially combine remaining pieces with result
+        for t in reversed(range(remaining_duration)):
+            result = prod_op(_shift_funsor(trans(**{time: t}), remaining_duration - t, global_vars), result)
+            sum_vars = frozenset(_shift_name(name, remaining_duration - t) for name in original_names)
+            result = result.reduce(sum_op, sum_vars)
+
+        result = result(**{name: _shift_name(name, -remaining_duration) for name in result.inputs})
+        return result
 
     for t in range(period):
         slice_t = Slice(time, t, duration - period + t + 1, period, duration)
-        factor = shift_funsor(trans, period - t - 1)
+        factor = _shift_funsor(trans, period - t - 1, global_vars)
         factor = factor(**{time: slice_t})
         renamed_factors.append(factor)
 
     block_trans = reduce(prod_op, renamed_factors)
-    block_step = {shift_name(name, period): name for name in block_trans.inputs
-                  if name != time and name not in global_vars and get_shift(name) < period}
+    block_step = {_shift_name(name, period): name for name in block_trans.inputs
+                  if name not in global_vars and _get_shift(name) < period}
     block_time_var = Variable(time_var.name, Bint[duration // period])
     final_chunk = mixed_sequential_sum_product(
         sum_op, prod_op, block_trans, block_time_var, block_step,
         num_segments=max(1, duration // (period * num_periods)))
     final_sum_vars = frozenset(
-        shift_name(name, t) for name in original_names for t in range(1, period))
+        _shift_name(name, t) for name in original_names for t in range(1, period))
     result = final_chunk.reduce(sum_op, final_sum_vars)
-    result = result(**{name: name.replace("P" * period, "P") for name in result.inputs})
+    result = result(**{name: _shift_name(name, -period + 1) for name in result.inputs})
     return result
 
 
@@ -385,7 +630,8 @@ class MarkovProduct(Funsor, metaclass=MarkovProductMeta):
                              if k != time.name)
         output = trans.output
         fresh = frozenset(step_names.values())
-        bound = frozenset(step_names.keys()) | {time.name}
+        bound = {k: trans.inputs[k] for k in step_names}
+        bound[time.name] = time.output
         super().__init__(inputs, output, fresh, bound)
         self.sum_op = sum_op
         self.prod_op = prod_op
@@ -395,7 +641,7 @@ class MarkovProduct(Funsor, metaclass=MarkovProductMeta):
         self.step_names = step_names
 
     def _alpha_convert(self, alpha_subs):
-        assert self.bound.issuperset(alpha_subs)
+        assert set(alpha_subs).issubset(self.bound)
         time = Variable(alpha_subs.get(self.time.name, self.time.name),
                         self.time.output)
         step = frozenset((alpha_subs.get(k, k), alpha_subs.get(v, v))
