@@ -333,6 +333,7 @@ def test_gaussian_mixture_distribution(batch_inputs, event_inputs):
     assert_close(q_marginal, p_marginal, atol=0.15, rtol=None)
 
 
+@pytest.mark.xfail(reason="numerically unstable")
 @pytest.mark.parametrize('moment', [0, 1, 2, 3])
 def test_lognormal_distribution(moment):
     num_samples = 100000
@@ -346,6 +347,7 @@ def test_lognormal_distribution(moment):
         with xfail_if_not_implemented():
             actual = Integrate(log_measure, probe, frozenset(['x']))
 
-    samples = backend_dist.LogNormal(loc, scale).sample((num_samples,))
+    _, (loc_data, scale_data) = align_tensors(loc, scale)
+    samples = backend_dist.LogNormal(loc_data, scale_data).sample((num_samples,))
     expected = (samples ** moment).mean(0)
     assert_close(actual.data, expected, atol=1e-2, rtol=1e-2)
